/* $Id$ */

/* SSLCipherCheck
 * Copyright (C) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "ciphers.h"
#include "tcp_common.h"

#define NAME "SSLCipherCheck"
#define VERSION "0.1"

#define FALSE 0
#define TRUE 1

#define MSGLEN 50
#define RESPONSELEN 2048
#define CIPHERSUITETEXTLEN 50

#define ARRAY_LEN(x)    sizeof(x) / sizeof(x[0])

enum {
  CIPHER_SUPPORTED = 10,
  CIPHER_NOT_SUPPORTED,
  ILLEGAL_RESPONSE,
  ERROR
};

/*
  TODO:
  - Add option for specifing specific cipher suite
  - Add option to print all ciphers defined in the program
  - Add option to brute force through all possible ciphers 0x0000 - 0xffff
*/

int create_clienthello(const struct cipher cipher_suite, unsigned char *msg);
void print_bytes(const unsigned char *ptr, const int len);
void get_ciphersuite_text(const int ssl_version,
                          const unsigned char cipher_suite[],
                          char *ciphersuitetext,
                          int ciphersuitetext_size);
int check_cipher(const char *hostname,
                 const int port,
                 const struct cipher cipher_suite);
void check_system(const char *hostname, const int port);
void check_ciphers(const char *hostname,
                   const int port,
                   const int protocol,
                   const struct cipher ciphers[],
                   const int array_siz);
void usage(const char *progname);
void help(const char *progname);

int verbose;


/*
 * main
 */
int main(int argc, char *argv[])
{
  char *hostname;
  int port;

  char *progname;

  int do_ssl2 = 0;
  int do_ssl3 = 0;
  int do_tls1 = 0;
  int do_system_check = 0;

  int c;

  verbose = 0;
  progname = argv[0];

  while(--argc > 0 && (*++argv)[0] == '-') {
    while ((c = *++argv[0])) {
      switch(c) {
      case 'v':
        verbose = 1;
        break;
      case '2':
        do_ssl2 = 1;
        break;
      case '3':
        do_ssl3 = 1;
        break;
      case 't':
        do_tls1 = 1;
        break;
      case 'c':
        do_system_check = 1;
        break;
      case 'h':
        help(progname);
        break;
      default:
        printf("%s: Illegal option %c\n", progname, c);
        printf("Try `%s -h' for more information.\n", progname);
        return EXIT_FAILURE;
      }
    }
  }

  if (argc != 2) {
    usage(progname);
    return EXIT_FAILURE;
  }
  
  hostname = argv[0];
  port = atoi(argv[1]); /* FIXME */

  if (do_system_check)
    check_system(hostname, port);
  if (do_ssl2)
    check_ciphers(hostname, port, SSL2, ssl2ciphers, ARRAY_LEN(ssl2ciphers));
  if (do_ssl3)
    check_ciphers(hostname, port, SSL3, ssl3ciphers, ARRAY_LEN(ssl3ciphers));
  if (do_tls1) 
    check_ciphers(hostname, port, TLS1, tls1ciphers, ARRAY_LEN(tls1ciphers));
  
  return EXIT_SUCCESS;
}


void usage(const char *progname)
{
  printf("Usage: %s [options] <hostname> <port>\n", progname);
}

void help(const char *progname)
{
  printf("%s %s, Copyright (c) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>\n"
         "\n"
         "Usage: %s [options] <hostname> <port>\n"
         "\n"
         "Options:\n"
         "   -2: Check SSL2 cipher suites\n"
         "   -3: Check SSL3 cipher suites\n"
         "   -t: Check TLS1 cipher suites\n"
         "   -c: Print information about remote system\n"
         "   -v: Be verbose\n"
         "   -h: Print this help text\n", NAME, VERSION, progname);
}



/*
 * Function to print out basic information about the host
 */
void check_system(const char *hostname, const int port) 
{
  /* hostname, port, webserver, OS... */
  printf("Information about remote system functin is not yet implemented\n");
}

/*
 * Function to check if a server support a set of ciphers.
 * This function also print out if the cipher actually was accepted or not.
 */
void check_ciphers(const char *hostname,
                   const int port,
                   const int protocol,
                   const struct cipher ciphers[],
                   const int cipher_array_siz)
{
  int i, ret;

  for (i = 0; i < cipher_array_siz; i++) {
    ret = check_cipher(hostname, port, ciphers[i]);
    if (ret == CIPHER_SUPPORTED) {
      printf("%s supported\n", ciphers[i].cipher);
    } else if (ret == CIPHER_NOT_SUPPORTED) {
      printf("%s NOT supported\n", ciphers[i].cipher);
    } else if (ret == ERROR) {
      perror("ERROR");
    } else {
      printf("Unexpected return value from check_cipher: %i\n", ret);
    }
  }
}

/* 
 * Function to check if a server support the specified cipher_suite.
 * returns CIPHER_SUPPORTED, CIPHER_NOT_SUPPORTED or ERROR if an error occured.
 */
int check_cipher(const char *hostname,
                 const int port,
                 const struct cipher cipher_suite)
{
  int socket;
  unsigned char msg[MSGLEN];  
  int msg_len;

  unsigned char response[RESPONSELEN];
  int bytes_read;

  int ssl_version;

  ssl_version = cipher_suite.protocol;

  if ((socket = tcp_connect(hostname, port)) < 0) {
    perror("ERROR in tcp_connect");
    return ERROR;
  }
  if ((msg_len = create_clienthello(cipher_suite, msg)) < 0) {
    printf("Error in create_clienthello\n");
    return ERROR;
  }
  if (verbose) {
    printf("DEBUG: ClientHello: ");
    print_bytes(msg, msg_len);
  }
  if (tcp_write_bytes(socket, msg, msg_len) != msg_len) {
    perror("ERROR in write_bytes");
    return ERROR;
  }
  if ((bytes_read = tcp_read_bytes(socket, response, RESPONSELEN)) < 0) {
    perror("Error in read");
    return ERROR;
  } else if (bytes_read == 0) {
    /* Some SSL server impementations will just close the
       connection, when no shared cipher is found */
    return CIPHER_NOT_SUPPORTED;
  }
  if (verbose) {
    printf("DEBUG: ServerHello: ");
    print_bytes(response, bytes_read);
  }
  if (close(socket) != 0) {
    perror("Error closing socket");
    return ERROR;
  }

  if (ssl_version == SSL3 || ssl_version == TLS1) {
    if (response[0] == 0x16) {
      /* We got a handshake message. Cipher suite was accepted */
      return CIPHER_SUPPORTED;
    } else if (response[0] == 0x15) {
      /* We got an alert message. Cipher not accepted by server */
      return CIPHER_NOT_SUPPORTED;
    } else {
      /* We got another type of message. This should not happen */
      print_bytes(response, bytes_read);
      return ERROR;
    }
  } else if (ssl_version == SSL2) {
    if (bytes_read >= 3 && response[0] == 0x00 && 
        response[1] == 0x00 && response[2] == 0x01) {
      /* We got SSL_MT_ERROR with SSL_PE_NO_CIPHER */
      return CIPHER_NOT_SUPPORTED;
    }
    else if (bytes_read >= 11 && response[2] == 0x04 &&
             response[9] == 0x00 && response[10] == 0x03) {
      /* We got ServerHello (0x04) and we got CipherSpec length == 1.
         Cipher suite was accepted */
      return CIPHER_SUPPORTED;
    } else if (response[9] == 0x00 && response[10] == 0x00) {
      /* We got CipherSpec length == 0 */
      return CIPHER_NOT_SUPPORTED;
    } else {
      /* We got another type of message. This should not happen */
      print_bytes(response, bytes_read);
      return ERROR;
    }
  } else {
    /* no such SSL version */
    return ERROR;
  }

}

/*
 * Function to create a SSL2/SSL3/TLS1 ClientHello handshake message. 
 * The output buffer msg must be at least 50 bytes. 
 * returns length of ClientHello message or -1 if an error occured.
 */
int create_clienthello(const struct cipher cipher_suite /* IN */,
                       unsigned char *msg /* OUT */ )
{
  unsigned char ssl3[] = { 
    /* SSL Record */
    0x16,                   /* type: Handshake */
    0x03, 0x00,             /* ProtocolVersion */
    0x00, 0x2d,             /* length: 45 */
    /* Handshake */
    0x01,                   /* ClientHello */
    0x00, 0x00, 0x29,       /* ClientHello length: 41 */
    0x03, 0x00,             /* ProtocolVersion */
    0x01, 0x02, 0x03, 0x04, /* Random.gmt_unix_time */
    0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x10,
    0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x20,
    0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, /* Random.random_bytes[28] */
    0x00,                   /* session_id length */
    0x00, 0x02,             /* CipherSuites length */
    0x00, 0x00,             /* CipherSuite */
    0x01,                   /* CompressionMethods length */
    0x00                    /* CompressionMethod: NULL */
  };

  unsigned char ssl2[] = {
    0x80, 0x1c, /* length: 28 */
    0x01, /* ClientHello */
    0x00, 0x02, /* Version: SSLv2 */
    0x00, 0x03, /* CipherSpec length: 3 */
    0x00, 0x00, /* SessionID length */
    0x00, 0x10, /* Challenge length: 16 */
    0x00, 0x00, 0x00, /* CipherSpec */
    0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x10,
    0x11, 0x12, 0x13, 0x14, 0x15, 0x16 /* Challenge */
  };

  assert(msg != NULL);

  if (cipher_suite.protocol == SSL2) {
    ssl2[11] = cipher_suite.bits[0];
    ssl2[12] = cipher_suite.bits[1];
    ssl2[13] = cipher_suite.bits[2];

    memcpy(msg, ssl2, sizeof(ssl2));
    return sizeof(ssl2);
  } 
  else {
    if (cipher_suite.protocol == SSL3)
      ssl3[2] = ssl3[10] = 0x00;
    else if (cipher_suite.protocol == TLS1)
      ssl3[2] = ssl3[10] = 0x01;
    else {
      printf("Unknown SSL version\n");  
      return -1;
    }
    ssl3[46] = cipher_suite.bits[0];
    ssl3[47] = cipher_suite.bits[1];

    memcpy(msg, ssl3, sizeof(ssl3));
    return sizeof(ssl3);
  } 
}

/* 
 * Function to print len number of bytes from ptr
 */
void print_bytes(const unsigned char *ptr, const int len)
{
  int i;

  for (i = 0; i < len; i++) {
    printf("0x%02x ", ptr[i]);
  }
  printf("\n");
}

