/* $Id$ */

/* SSLResumeCheck
 * Copyright (C) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <openssl/err.h>
#include <openssl/ssl.h>

#include "tcp_common.h"

#define NAME "SSLResumeCheck"
#define VERSION "0.1"

#define int_error(msg)  handle_error(__FILE__, __LINE__, msg)

void check_resume(const char *hostname,
                  const int port, 
                  const int protocol);
void init_OpenSSL(void);
void handle_error(const char *file, int lineno, const char *msg);
SSL_CTX *setup_client_ctx(const int protocol);
void usage(const char *progname);
void help(const char *progname);
void print_bytes(const unsigned char *ptr, const int len);

int equal_session_id(const unsigned char *session_id1,
                     const int session_id1_length,
                     const unsigned char *session_id2,
                     const int session_id2_length);

enum protocols {
  SSL2 = 0,
  SSL3,
  TLS1
};

int verbose;

int main(int argc, char *argv[])
{
  char *progname;
  char *hostname;
  int port;

  int do_ssl2 = 0;
  int do_ssl3 = 1;
  int do_tls1 = 0;

  int c;

  verbose = 0;
  progname = argv[0];

  while(--argc > 0 && (*++argv)[0] == '-') {
    while ((c = *++argv[0])) {
      switch(c) {
      case 'v':
        verbose = 1;
        break;
      case '2':
        do_ssl2 = 1;
        do_ssl3 = 0;
        do_tls1 = 0;
        break;
      case '3':
        do_ssl2 = 0;
        do_ssl3 = 1;
        do_tls1 = 0;
        break;
      case 't':
        do_ssl2 = 0;
        do_ssl3 = 0;
        do_tls1 = 1;
        break;
      case 'h':
        help(progname);
        break;
      default:
        printf("%s: Illegal option %c\n", progname, c);
        printf("Try `%s -h' for more information.\n", progname);
        return EXIT_FAILURE;
      }
    }
  }

  if (argc != 2) {
    usage(progname);
    return EXIT_FAILURE;
  }
    
  hostname = argv[0];
  port = atoi(argv[1]); /* FIXME */

  if (do_ssl2)
    check_resume(hostname, port, SSL2);
  else if (do_ssl3)
    check_resume(hostname, port, SSL3);
  else if (do_tls1) 
    check_resume(hostname, port, TLS1);
  else
    fprintf(stderr, "Invalid do_ssl variable\n");

  return EXIT_SUCCESS;
}

void usage(const char *progname)
{
  printf("Usage: %s [options] <hostname> <port>\n", progname);
}

void help(const char *progname)
{
  printf("%s %s, Copyright (c) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>\n"
         "\n"
         "Usage: %s [options] <hostname> <port>\n"
         "\n"
         "Options:\n"
         "   -2: Use SSL2 resume\n"
         "   -3: Use SSL3 resume (default)\n"
         "   -t: Use TLS1 resume\n"
         "   -v: Be verbose\n"
         "   -h: Print this help text\n", NAME, VERSION, progname);
}

void check_resume(const char *hostname,
                  const int port, 
                  const int protocol)
{
  SSL *ssl, *ssl_resumed;
  SSL_SESSION *saved_session1, *saved_session2;
  SSL_CTX *ctx;
  int socket, socket_resumed;
  char *version;
  int session_id1_length;
  unsigned char session_id1[SSL_MAX_SSL_SESSION_ID_LENGTH];
  int session_id2_length;
  unsigned char session_id2[SSL_MAX_SSL_SESSION_ID_LENGTH];


  switch (protocol){
  case SSL2:
    version = "SSL2";
    break;
  case SSL3:
    version = "SSL3";
    break;
  case TLS1:
    version = "TLS1";
    break;
  default:
    fprintf(stderr, "Invalid version: %i\n", protocol);
    exit(-1);
  }

  printf("Checking if host %s:%i accept %s session"
         "resumption after TCP FIN\n", hostname, port, version);

  init_OpenSSL();
  /* seed_prng() */

  /* create context to hold certificates, algorithms, protocol versions etc. */
  ctx = setup_client_ctx(protocol);

  /* create an SSL object */
  if (!(ssl = SSL_new(ctx)))
    int_error("Error creating SSL context");

  /* create TCP socket */
  if ((socket = tcp_connect(hostname, port)) <= 0)
    int_error("Error creating socket");

  /* associate a network connection to the SSL object */
  /* SSL_set_fd() or SSL_set_bio() */
  if (!SSL_set_fd(ssl, socket))
    int_error("Error associating network connection to SSL object");

  /* perform handshake */
  if (SSL_connect(ssl) <= 0)
    int_error("Error connecting SSL object");

  /* save session */
  if ((saved_session1 = SSL_get1_session(ssl)) == NULL)
    int_error("Error saving SSL session");

  /* save the session id */
  session_id1_length = saved_session1->session_id_length;
  memcpy(session_id1, saved_session1->session_id, session_id1_length);

  printf("SSL session opened with session_id = \n");
  print_bytes(session_id1, session_id1_length);
  printf("Sleeping for 5 sec. before closing SSL session with TCP FIN....\n");
  sleep(5);

  /* close TCP socket */
  if (tcp_close(socket) != 0)
    int_error("Error closing TCP socket");
  printf("Sent TCP FIN to server.\n");

  printf("Sleeping for 5 sec. before resuming session....\n");
  sleep(5);
  


  /* start to resume the session */

  /* create a new SSL object */
  if (!(ssl_resumed = SSL_new(ctx)))
    int_error("Error creating resumed SSL context");

  /* create a new TCP socket */
  if ((socket_resumed = tcp_connect(hostname, port)) <= 0)
    int_error("Error creating resumed socket");

  /* associate a network connection to the SSL object */
  if (!SSL_set_fd(ssl_resumed, socket_resumed))
    int_error("Error associating network connection to resumed SSL object");

  /* resume old session */
  if (SSL_set_session(ssl_resumed, saved_session1) != 1)
    int_error("Error setting resumed session");

  /* free buffer of saved session */
  SSL_SESSION_free(saved_session1);

  /* connect using resumed session*/
  if (SSL_connect(ssl_resumed) <= 0)
    int_error("Error connecting resumed SSL object");

  /* save the new resumed session */
  if ((saved_session2 = SSL_get1_session(ssl_resumed)) == NULL)
    int_error("Error saving SSL session2");

  /* save the session_id of the new resumed session */
  session_id2_length = saved_session2->session_id_length;
  memcpy(session_id2, saved_session2->session_id, session_id2_length);  

  printf("SSL session successfully established with session_id = \n");
  print_bytes(session_id2, session_id2_length);

  /* check if the session id's are equal */
  if (equal_session_id(session_id1, session_id1_length, session_id2, session_id2_length))
    printf("The two session_id's are equal. This is a violation of the standard!\n");
  else
    printf("The server generated a new session_id. This is good!\n");

  SSL_shutdown(ssl_resumed);
  SSL_free(ssl_resumed);
}

/*
  returns 0 if the two session_id's are not equal in bytes and length
  returns 1 if they are equal.
*/
int equal_session_id(const unsigned char *session_id1,
                     const int session_id1_length,
                     const unsigned char *session_id2,
                     const int session_id2_length)
{
  int i;

  if (session_id1_length != session_id2_length)
    return 0;

  if (session_id1_length == 0)
    return 0;

  for (i=0; i < session_id1_length; i++) {
    if (session_id1[i] != session_id2[i])
      return 0;
  }

  return 1;
}

/* 
 * Function to print len number of bytes from ptr
 */
void print_bytes(const unsigned char *ptr, const int len)
{
  int i;

  for (i = 0; i < len; i++) {
    printf("0x%02x ", ptr[i]);
  }
  printf("\n");
}

SSL_CTX *setup_client_ctx(const int protocol)
{
  SSL_CTX *ctx;

  /* SSLv2_client_method(), SSLv3_client_method(), TLSv1_client_method() */
  switch(protocol) {
  case SSL2:
    ctx = SSL_CTX_new(SSLv2_client_method());
    break;
  case SSL3:
    ctx = SSL_CTX_new(SSLv3_client_method());
    break;
  case TLS1:
    ctx = SSL_CTX_new(TLSv1_client_method());
    break;
  default:    
    fprintf(stderr, "** Invalid protocol: %i\n", protocol);
    exit(-1);
  }

  return ctx;
}

void init_OpenSSL(void) 
{
  if (!SSL_library_init()) {
    fprintf(stderr, "** OpenSSL initialization failed!\n");
    exit(-1);
  }
  SSL_load_error_strings();
}

void handle_error(const char *file, int lineno, const char *msg)
{
  fprintf(stderr, "** %s:%i %s\n", file, lineno, msg);
  ERR_print_errors_fp(stderr);
  exit(-1);
}

