/* $Id$ */

/* SSLVersionCheck
 * Copyright (C) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdlib.h>
#include <string.h>

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/rand.h>

#include "tcp_common.h"

#define NAME "SSLVersionCheck"
#define VERSION "0.1"

#define int_error(msg)  handle_error(__FILE__, __LINE__, msg)

char *read_http_head(SSL *ssl);
void init_OpenSSL(void);
int get_version(char *server, int port);
int write_http_head(SSL *ssl);
char *parse_http_head(char *head);
void usage(const char *progname);
void help(const char *progname);
void handle_error(const char *file, int lineno, const char *msg);

int verbose;

int main(int argc, char *argv[])
{
  char *progname;
  char *hostname;
  int port;
  int c;

  verbose = 0;
  progname = argv[0];

  while(--argc > 0 && (*++argv)[0] == '-') {
    while ((c = *++argv[0])) {
      switch(c) {
      case 'v':
        verbose = 1;
        break;
      case 'h':
        help(progname);
        break;
      default:
        printf("%s: Illegal option %c\n", progname, c);
        printf("Try `%s -h' for more information.\n", progname);
        return EXIT_FAILURE;
      }
    }
  }

  if (argc != 2) {
    usage(progname);
    return EXIT_FAILURE;
  }
    
  hostname = argv[0];
  port = atoi(argv[1]); /* FIXME */

  get_version(hostname, port);

  return EXIT_SUCCESS;
}

void usage(const char *progname)
{
  printf("Usage: %s [options] <hostname> <port>\n", progname);
}

void help(const char *progname)
{
  printf("%s %s, Copyright (c) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>\n"
         "\n"
         "Usage: %s [options] <hostname> <port>\n"
         "\n"
         "Options:\n"
         "   -v: Be verbose\n"
         "   -h: Print this help text\n", NAME, VERSION, progname);
}

int get_version(char *hostname, int port)
{
  SSL *ssl;
  SSL_CTX *ctx;
  int socket;
  char *response;
  char *ssl_version;

  init_OpenSSL();
  RAND_load_file("/dev/random", 1024);
  ctx = SSL_CTX_new(SSLv23_method());

  SSL_CTX_set_options(ctx, SSL_OP_ALL);

  if (!(ssl = SSL_new(ctx)))
    int_error("Error creating SSL object");
  if ((socket = tcp_connect(hostname, port)) <= 0)
    int_error("Error creating socket");
  if (!SSL_set_fd(ssl, socket))
    int_error("Error associating network connection to SSL object");
  if (SSL_connect(ssl) <= 0)
    int_error("Error connecting SSL object");

  if (verbose)
    printf("Connected\n");
  write_http_head(ssl);
  if (verbose)
    printf("Wrote HEAD\n");
  response = read_http_head(ssl);
  if (verbose)
    printf("Response: %s\n", response);
  ssl_version = parse_http_head(response);

  printf("%s: %s\n", hostname, ssl_version);

  return 0;
}

char *parse_http_head(char *head)
{
  char *buf;
  char *pos1, *pos2;
  char *ptr;
  int len;
  char *server_string = "\nServer:";

  pos1 = strstr(head, server_string);
  if (pos1 == NULL)
    int_error("Server string not found in answer from webserver\n");

  ptr = pos1 + strlen(server_string);

  pos2 = strchr(ptr, '\n');
  if (pos2 == NULL)
    int_error("No newline after Server string\n");

  len = pos2 - ptr;

  buf = malloc(len + 1);
  if (buf == NULL)
    int_error("Malloc failed\n");
  memcpy(buf, ptr, len);
  buf[len] = '\0';

  return buf;
}

int write_http_head(SSL *ssl)
{
  int err;
  const char* head = "HEAD / HTTP/1.0\n\n\n\n";

  err = SSL_write(ssl, head, strlen(head));
  if (err <= 0)
    return 0;
  return 1;
}

char *read_http_head(SSL *ssl)
{
  int err, nread;
  char buf[1024];
  char *response;

  for (nread = 0; nread < sizeof(buf); nread += err) {
    err = SSL_read(ssl, buf + nread, sizeof(buf) - nread);
    if (err <= 0)
      break;
  }

  response = malloc(nread+1);
  if (response == NULL)
    int_error("Malloc failed!\n");

  memcpy(response, buf, nread);
  response[nread] = '\0';

  return response;
}

void init_OpenSSL(void)
{
  if(!SSL_library_init())
    fprintf(stderr, "OpenSSL initialization failed!\n");
  
  SSL_load_error_strings();
}

void handle_error(const char *file, int lineno, const char *msg)
{
  fprintf(stderr, "** %s:%i %s\n", file, lineno, msg);
  ERR_print_errors_fp(stderr);
  exit(-1);
}
