/* $Id$ */

/*
 * Copyright (C) 2003 Jostein Tveit <josteitv@pvv.ntnu.no>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "tcp_common.h"

/*
 * Function to make a TCP connection to port on host.
 * returns the socket descriptor, or -1 if an error occured.
 */
int tcp_connect(const char *host, const uint16_t port)
{
  struct hostent *hp;
  struct sockaddr_in addr;
  int sock;

  if (!(hp=gethostbyname(host))) {
    perror("ERROR: gethostbyname() failed");
    return -1;
  }

  memset(&addr, 0, sizeof(addr));

  addr.sin_addr=*(struct in_addr*)hp->h_addr_list[0];
  addr.sin_family=AF_INET;
  addr.sin_port=htons(port);

  if((sock=socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    perror("ERROR: Could not create socket");
    
    return -1;
  }
  if(connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
    perror("ERROR: Could not connect to socket");
    return -1;
  }

  return sock;
}

/*
 * Function to close a TCP socket.
 * Just a wrapper around close(2).
 * returns 0 on success, -1 if en error occured.
 */
int tcp_close(int fd)
{
  int ret;

  if ((ret = close(fd)) != 0)
    perror("ERROR: Could not close socket");
  
  return ret;
}


int tcp_listen(const uint16_t port, const int backlog)
{
  struct sockaddr_in addr;
  int sock, connfd;

  memset(&addr, 0, sizeof(addr));

  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(port);

  if((sock=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("ERROR: Could not connect to socket");
    return -1;
  }

  if (bind(sock, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    perror("ERROR: Could not bind to socket");
    return -1;
  }

  if (listen(sock, backlog) < 0) {
    perror("ERROR: Could not listen to socket");
    return -1;
  }
  
  if((connfd = accept(sock, NULL, NULL)) < 0) {
    perror("ERROR: Could not accept socket");
    return -1;
  }

  return connfd;
}


/*
 * Write len number of bytes to socket.
 * returns number of bytes written to socket or -1 if an error occured.
 */
int tcp_write_bytes(const int socket, const void *bytes, const size_t len)
{
  size_t bytes_left;
  ssize_t just_written;
  const char *ptr;

  ptr = bytes;
  bytes_left = len;
  while (bytes_left > 0) {
    if ((just_written = write(socket, ptr, bytes_left)) <= 0) {
      if (errno == EINTR)
        just_written = 0;
      else
        return -1;
    }
    bytes_left -= just_written;
    ptr += just_written;
  }

  /* function will always write len bytes */
  return len; 
}


/* 
 * Function to read from socket.
 * returns number of bytes read or -1 if an error occured. 
 */
int tcp_read_bytes(const int socket /* IN */, 
               unsigned char *response /* OUT */,
               int response_size /* IN */)
{
  int cnt;

  cnt = read(socket, response, response_size);
  if (cnt < 0) {
    return -1;
  }
  
  return cnt;
}
